import pyjags
import numpy as np

np.random.seed(0)
np.set_printoptions(precision=1)

N = 500
a = 70
b = 4
sigma = 50

x = np.random.uniform(0, 100, size=N)
y = np.random.normal(a + x*b, sigma, size=N)

code = '''
model {
    for (i in 1:N) {
        y[i] ~ dnorm(alpha + beta * x[i], tau)
    }
    alpha ~ dunif(-1e3, 1e3)
    beta ~ dunif(-1e3, 1e3)
    tau <- 1 / sigma^2
    sigma ~ dgamma(1e-4, 1e-4)
}
'''

model = pyjags.Model(code, data=dict(x=x, y=y, N=N), chains=4)
samples = model.sample(5000, vars=['alpha', 'beta', 'sigma'])

def summary(samples, varname, p=95):
    values = samples[varname]
    ci = np.percentile(values, [100-p, p])
    print('{:<6} mean = {:>5.1f}, {}% credible interval [{:>4.1f} {:>4.1f}]'.format(
      varname, np.mean(values), p, *ci))

for varname in ['alpha', 'beta', 'sigma']:
    summary(samples, varname)



# Copyright (C) 2016 Tomasz Miasko
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

__all__ = ['version', 'get_modules_dir', 'set_modules_dir', 'list_modules', 'load_module', 'unload_module']

import ctypes
import ctypes.util
import os
import logging
import sys

from .console import Console

logger = logging.getLogger('pyjags')
modules_dir = None


def version():
    """JAGS version as a tuple of ints.

    >>> pyjags.version()
    (3, 4, 0)
    """
    v = Console.version()
    return tuple(map(int, v.split('.')))


if sys.platform.startswith('darwin'):
    def list_shared_objects():
        """Return paths of all currently loaded shared objects."""

        libc = ctypes.util.find_library('c')
        libc = ctypes.cdll.LoadLibrary(libc)

        dyld_image_count = libc._dyld_image_count
        dyld_image_count.argtypes = []
        dyld_image_count.restype = ctypes.c_uint32

        dyld_image_name = libc._dyld_get_image_name
        dyld_image_name.argtypes = [ctypes.c_uint32]
        dyld_image_name.restype = ctypes.c_char_p

        libraries = []

        for index in range(dyld_image_count()):
            libraries.append(dyld_image_name(index))

        return list(map(getattr(os, 'fsdecode', lambda x: x), libraries))

elif sys.platform.startswith('linux'):
    def list_shared_objects():
        """Return paths of all currently loaded shared objects."""

        class dl_phdr_info(ctypes.Structure):
            _fields_ = [
                ('addr', ctypes.c_void_p),
                ('name', ctypes.c_char_p),
                ('phdr', ctypes.c_void_p),
                ('phnum', ctypes.c_uint16),
            ]

        dl_iterate_phdr_callback = ctypes.CFUNCTYPE(
                ctypes.c_int,
                ctypes.POINTER(dl_phdr_info),
                ctypes.POINTER(ctypes.c_size_t),
                ctypes.c_void_p)

        libc = ctypes.util.find_library('c')
        libc = ctypes.cdll.LoadLibrary(libc)
        dl_iterate_phdr = libc.dl_iterate_phdr
        dl_iterate_phdr.argtypes = [dl_iterate_phdr_callback, ctypes.c_void_p]
        dl_iterate_phdr.restype = ctypes.c_int

        libraries = []

        def callback(info, size, data):
            path = info.contents.name
            if path:
                libraries.append(path)
            return 0

        dl_iterate_phdr(dl_iterate_phdr_callback(callback), None)

        return list(map(getattr(os, 'fsdecode', lambda x: x), libraries))

else:
    def list_shared_objects():
        """Return paths of all currently loaded shared objects."""
        return []


def locate_modules_dir_using_shared_objects():
    for path in list_shared_objects():
        name = os.path.basename(path)
        if name.startswith('jags') or name.startswith('libjags'):
            dir = os.path.dirname(path)
            logger.info('Using JAGS library located in %s.', path)
            return os.path.join(dir, 'JAGS', 'modules-{}'.format(version()[0]))
    return None


def locate_modules_dir():
    logger.debug('Locating JAGS module directory.')
    return locate_modules_dir_using_shared_objects()


def get_modules_dir():
    """Return modules directory."""
    global modules_dir
    if modules_dir is None:
        modules_dir = locate_modules_dir()
    if modules_dir is None:
        raise RuntimeError(
            'Could not locate JAGS module directory. Use pyjags.set_modules_dir(path) to configure it manually.')
    return modules_dir


def set_modules_dir(directory):
    """Set modules directory."""
    global modules_dir
    modules_dir = directory


def list_modules():
    """Return a list of loaded modules."""
    return Console.listModules()


def load_module(name, modules_dir=None):
    """Load a module.

    Parameters
    ----------
    name : str
        A name of module to load.
    modules_dir : str, optional
        Directory where modules are located.
    """
    if name not in loaded_modules:
        dir = modules_dir or get_modules_dir()
        ext = '.so' if os.name == 'posix' else '.dll'
        path = os.path.join(dir, name + ext)
        logger.info('Loading module %s from %s', name, path)
        module = ctypes.cdll.LoadLibrary(path)
        loaded_modules[name] = module
    Console.loadModule(name)

loaded_modules = {}


def unload_module(name):
    """Unload a module."""
    return Console.unloadModule(name)



# Copyright (C) 2015-2016 Tomasz Miasko
#               2020 Michael Nowotny
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

__all__ = ['Model']

import collections
import contextlib
import sys
import tempfile

import numpy as np

from .console import Console, DUMP_ALL, DUMP_DATA, DUMP_PARAMETERS
from .modules import load_module
from .progressbar import const_time_partition, progress_bar_factory

# Special value indicating missing data in JAGS.
JAGS_NA = -sys.float_info.max*(1-1e-15)


def dict_to_jags(src):
    """Convert Python dictionary with array like values to format suitable
    for use with JAGS.

     * Returned arrays have at least one dimension.
     * Empty arrays are removed from the dictionary.
     * Masked values are replaced with JAGS_NA.
    """
    dst = {}
    for k, v in src.items():
        if np.ma.is_masked(v):
            v = np.ma.array(data=v, dtype=np.double, ndmin=1,
                            fill_value=JAGS_NA)
            v = np.ma.filled(v)
        else:
            v = np.atleast_1d(v)
        if not np.size(v):
            continue
        dst[k] = v
    return dst


def dict_from_jags(src):
    """Convert Python dictionary with array like values returned from JAGS to
    format suitable for use with Python.

     * Arrays containing JAGS_NA values are converted to numpy MaskedArray.
    """
    dst = {}
    for k, v in src.items():
        mask = v == JAGS_NA
        # Don't mask if it not necessary
        if np.any(mask):
            v = np.ma.masked_equal(v, JAGS_NA, copy=False)
        dst[k] = v
    return dst


def check_locale_compatibility():
    """Checks that current locale is compatible with JAGS."""
    import locale
    import textwrap

    if locale.localeconv().get('decimal_point') != '.':
        msg = """\
        JAGS requires locale that uses period as decimal point character. The standard C locale would be one possible choice. It can be configured as follows:
        > import locale
        > locale.setlocale(locale.LC_ALL, 'C')"""
        raise ValueError(textwrap.dedent(msg))


@contextlib.contextmanager
def model_path(file=None, code=None, encoding='utf-8'):
    """Utility function returning model path, if necessary creates a
    new temporary file with a model code written into it.
    """
    if file:
        yield file
    elif code:
        if isinstance(code, str):
            code = code.encode(encoding=encoding)
        # TODO use separate delete to support Windows?
        with tempfile.NamedTemporaryFile() as fh:
            fh.write(code)
            fh.flush()
            yield fh.name
    else:
        raise ValueError('Either model name or model text must be provided.')


class MultiConsole:

    def __init__(self, chains, chains_per_thread):
        # Multiple consoles that emulate a single JAGS console.
        self.consoles = []
        self.chains_per_console = []
        # Map from outer chain number to inner console and its inner chain number.
        # Uses JAGS indexing from 1.
        self.chains = {}

        outer_chain = 1
        while chains > 0:
            console = Console()
            console_chains = min(chains_per_thread, chains)

            self.consoles.append(console)
            self.chains_per_console.append(console_chains)

            for inner_chain in range(1, console_chains+1):
                self.chains[outer_chain] = (console, inner_chain)
                outer_chain += 1

            chains -= chains_per_thread

    def checkModel(self, path):
        for c in self.consoles:
            c.checkModel(path)

    def compile(self, data, chains, generate_data):
        assert(chains == len(self.chains))
        for console, chains in zip(self.consoles, self.chains_per_console):
            console.compile(data, chains, generate_data)

    def setRNGname(self, name, chain):
        console, chain = self.chains[chain]
        console.setRNGname(name, chain)

    def setParameters(self, data, chain):
        console, chain = self.chains[chain]
        console.setParameters(data, chain)

    def setMonitor(self, name, thin, type):
        for c in self.consoles:
            c.setMonitor(name, thin, type)

    def setMonitors(self, names, thin, type):
        for name in names:
            self.setMonitor(name, thin, type)

    def clearMonitor(self, name, type):
        for c in self.consoles:
            c.clearMonitor(name, type)

    def dumpMonitors(self, type, flat):
        ds = [c.dumpMonitors(type, flat) for c in self.consoles]
        return {k: np.concatenate([d[k] for d in ds], axis=-1)
                for k in set(k for d in ds for k in d.keys())}

    def initialize(self):
        for c in self.consoles:
            c.initialize()

    def isAdapting(self):
        return any(c.isAdapting() for c in self.consoles)

    def checkAdaptation(self):
        return any(c.checkAdaptation() for c in self.consoles)

    def variableNames(self):
        return self.consoles[0].variableNames()

    def dumpState(self, type, chain):
        console, chain = self.chains[chain]
        return console.dumpState(type, chain)


class Model:
    """High level representation of JAGS model.

    Attributes
    ----------
    chains : int
        A number of chains in the model.

    Note
    ----
    In JAGS arrays are indexed from 1. On the other hand Python uses 0 based
    indexing. It is important to keep this in mind when providing data to JAGS
    and interpreting resulting samples. For example, what in JAGS would be
    x[4,2,7] in Python is x[3,1,6].

    Note
    ----
    The JAGS supports data sets where some of observations have no value.
    In PyJAGS those missing values are described using numpy MaskedArray.
    For example, to create a model with observations x[1] = 0.25, x[3] = 0.75,
    and observation x[2] missing, we would provide following data to Model
    constructor:

    >>> {'x': np.ma.masked_array(data=[0.25, 0, 0.75], mask=[False, True, False])}
    {'x': masked_array(data = [0.25 -- 0.75],
                 mask = [False  True False],
           fill_value = 1e+20)
    }

    From JAGS version 4.0.0 it is also possible to monitor variables that are
    not completely defined in the description of the model, e.g., if y[i] is
    defined only for y[3], then y[1], and y[2] will have missing values for
    all iterations in all chains. Those missing values are also represented
    using numpy MaskedArray.
    """

    def __init__(self, code=None, data=None, init=None, chains=4, adapt=1000,
                 file=None, encoding='utf-8', generate_data=True,
                 progress_bar=True, refresh_seconds=None,
                 threads=1, chains_per_thread=1):
        """
        Create a JAGS model and run adaptation steps.

        Parameters
        ----------
        code : str or bytes, optional
            Code of the model to load. Model may be also provided with file
            keyword argument.
        file : str, optional
            Path to the model to load. Model may be also provided with code
            keyword argument.
        init : dict or list of dicts, optional
            Specifies initial values for parameters. It can be either a
            dictionary providing initial values for parameters used as keys,
            or a list of dictionaries providing initial values separately for
            each chain. If omitted, initial values will be generated
            automatically.

            Additionally this option allows to configure random number
            generators using following special keys:

             * '.RNG.name'  str, name of random number generator
             * '.RNG.seed'  int, seed for random number generator
             * '.RNG.state' array, may be specified instead of seed, shape of
               array depends on particular generator used
        data : dict, optional
            Dictionary with observed nodes in the model. Keys are variable
            names and values should be convertible to numpy arrays with shape
            compatible with one used in the model.

            The numpy.ma.MaskedArray can be used to provide data where some of
            observations are missing.
        generate_data : bool, optional
            If true, data block in the model is used to generate data.
        chains : int, 4 by default
            A positive number specifying number of parallel chains.
        adapt : int, 1000 by default
            An integer specifying number of adaptations steps.
        encoding : str, 'utf-8' by default
            When model code is provided as a string, this specifies its encoding.
        progress_bar : bool, optional
            If true, enables the progress bar.
        threads: int, 1 by default
            A positive integer specifying number of threads used to sample from
            model. Using more than one thread is experimental functionality.
        chains_per_thread: int, 1 by default
            A positive integer specifying a maximum number of chains sampled in
            a single thread. Takes effect only when using more than one thread.
        """

        check_locale_compatibility()

        # Ensure that default modules are loaded.
        load_module('basemod')
        load_module('bugs')
        load_module('lecuyer')

        self.refresh_seconds = refresh_seconds or 0.5 if sys.stdout.isatty() else 5.0
        self.progress_bar = progress_bar_factory(progress_bar, refresh_seconds=self.refresh_seconds)
        self.chains = chains
        self.threads = threads
        self.use_threads = self.threads > 1 and chains_per_thread < self.chains

        if self.use_threads:
            self.console = MultiConsole(self.chains, chains_per_thread)
        else:
            self.console = Console()

        with model_path(file, code, encoding) as path:
            self.console.checkModel(path)

        self._init_compile(data, generate_data)
        self._init_parameters(init)
        self.console.initialize()
        if adapt:
            self.adapt(adapt)

    def _init_compile(self, data, generate_data):
        if data is None:
            data = {}
        data = dict_to_jags(data)
        unused = set(data.keys()) - set(self.variables)
        if unused:
            raise ValueError(
                'Unused data for variables: {}'.format(','.join(unused)))
        self.console.compile(data, self.chains, generate_data)

    def _init_parameters(self, init):
        """Set parameters and configure random number generators."""
        if init is None:
            init = {}
        if isinstance(init, collections.Mapping):
            init = [init] * self.chains
        elif not isinstance(init, collections.Sequence):
            raise ValueError('Init should be a sequence or a dictionary.')
        if len(init) != self.chains:
            raise ValueError(
                'Length of init sequence should equal the number of chains.')

        if self.use_threads:
            rngs = Console.parallel_rngs('lecuyer::RngStream', self.chains)
        else:
            rngs = [{'.RNG.name': None, '.RNG.seed': None}] * self.chains

        for data, rng, chain in zip(init, rngs, range(1, self.chains + 1)):
            data = dict(data)
            rng_name = data.pop('.RNG.name', None)
            if self.use_threads and rng_name is None:
                rng_name = rng['.RNG.name']
                data['.RNG.state'] = rng['.RNG.state']
            if rng_name is not None:
                self.console.setRNGname(rng_name, chain)
            data = dict_to_jags(data)

            unused = set(data.keys())
            unused.difference_update(self.variables)
            unused.difference_update(['.RNG.seed', '.RNG.state'])
            if unused:
                raise ValueError(
                    'Unused initial values in chain {} for variables: {}'.format(
                        chain, ','.join(unused)))
            self.console.setParameters(data, chain)

    def _update(self, iterations, header):
        if self.use_threads:
            method = self._update_parallel
        else:
            method = self._update_sequential

        with self.progress_bar(self.chains * iterations, header=header) as pb:
            method(pb, iterations)

    def _update_sequential(self, progress, iterations):
        for steps in const_time_partition(iterations, self.refresh_seconds):
            self.console.update(steps)
            progress.update(self.chains * steps)

    def _update_parallel(self, progress, iterations):
        from concurrent.futures import ALL_COMPLETED, ThreadPoolExecutor, wait
        from threading import Event

        with ThreadPoolExecutor(self.threads) as executor:
            # Event used to interrupt inner threads (which are
            # non-interruptable by default).
            interrupt = Event()

            def update(console, chains):
                for steps in const_time_partition(iterations, self.refresh_seconds):
                    if interrupt.is_set():
                        break
                    console.update(steps)
                    progress.update(chains * steps)
            fs = [executor.submit(update, console, chains)
                  for console, chains in zip(self.console.consoles,
                                             self.console.chains_per_console)]
            try:
                (done, not_done) = wait(fs, return_when=ALL_COMPLETED)
                for d in done:
                    d.result()
                for d in not_done:
                    assert(False)
            except KeyboardInterrupt:
                interrupt.set()
                raise

    def update(self, iterations):
        """Updates the model for given number of iterations."""
        self._update(iterations, 'updating: ')

    def sample(self, iterations, vars=None, thin=1, monitor_type="trace"):
        """
        Creates monitors for given variables, runs the model for provided
        number of iterations and returns monitored samples.


        Parameters
        ----------
        iterations : int
            A positive integer specifying number of iterations.
        vars : list of str, optional
            A list of variables to monitor.
        thin : int, optional
            A positive integer specifying thinning interval.
        Returns
        -------
        dict
            Sampled values of monitored variables as a dictionary where keys
            are variable names and values are numpy arrays with shape:
            (dim_1, dim_n, iterations, chains). dim_1, ..., dim_n describe the
            shape of variable in JAGS model.
        """
        if vars is None:
            vars = self.variables
        monitored = []
        try:
            for name in vars:
                self.console.setMonitor(name, thin, monitor_type)
                monitored.append(name)
            self._update(iterations, 'sampling: ')
            samples = self.console.dumpMonitors(monitor_type, False)
            samples = dict_from_jags(samples)
        finally:
            for name in monitored:
                self.console.clearMonitor(name, monitor_type)
        return samples

    def adapt(self, iterations):
        """Run adaptation steps to maximize samplers efficiency.

        Returns
        -------
        bool
            True if achieved performance is close to the theoretical optimum.
        """
        if not self.console.isAdapting():
            # Model does not require adaptation
            return True
        self._update(iterations, 'adapting: ')
        return self.console.checkAdaptation()

    @property
    def variables(self):
        """Variable names used in the model."""
        return self.console.variableNames()

    @property
    def state(self):
        """Values of model parameters and model data for each chain.

        See Also
        --------
        parameters
        data
        """
        return [dict_from_jags(self.console.dumpState(DUMP_ALL, chain))
                for chain in range(1, self.chains + 1)]

    @property
    def parameters(self):
        """Values of model parameters for each chain. Includes name of random
        number generator as '.RNG.name' and its state as '.RNG.state'.
        """
        return [dict_from_jags(self.console.dumpState(DUMP_PARAMETERS, chain))
                for chain in range(1, self.chains + 1)]

    @property
    def data(self):
        """Model data. Includes data provided during model construction and
        data generated as part of data block.
        """
        return dict_from_jags(self.console.dumpState(DUMP_DATA, 1))



# Copyright (C) 2015-2016 Tomasz Miasko
#               2020 Michael Nowotny
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import numbers
import numpy as np

from .modules import load_module
from .model import Model


class DiffDIC:
    def __init__(self, delta):
        """
        This class stores the difference in the DIC between two models.

        Parameters
        ----------
        delta: Numpy array with the difference in the DIC for each observations
        """
        self._delta = delta

        if isinstance(self.delta, np.ndarray):
            self._n = len(self.delta)
        elif isinstance(self.delta, numbers.Number):
            self._n = 1
        else:
            raise TypeError(f'delta must either be a numpy array or a number '
                            f'but is of type {type(delta)}')

    @property
    def delta(self):
        return self._delta

    def __str__(self):

        result = f"Difference: {np.sum(self.delta)}\n"
        result += f"Sample standard error: " \
                  f"{np.sqrt(self._n) * np.std(self.delta)}"
        return result

    def __repr__(self):
        return self.__str__()


class DIC:
    def __init__(self, deviance, penalty, type):
        """
        This class stores the Deviance Information Criterion (DIC) sampled from
        a PyJAGS model.

        Parameters
        ----------
        deviance
        penalty
        type: either 'pD' or 'popt'
        """
        self._deviance = deviance
        self._penalty = penalty
        self._type = type

    @property
    def deviance(self):
        return self._deviance

    @property
    def penalty(self):
        return self._penalty

    @property
    def type(self):
        return self._type

    def construct_report(self, digits = 2) -> str:
        result = ""
        deviance = np.sum(self.deviance)
        psum = np.sum(self.penalty)
        result += "Mean deviance: {:.{}f}\n".format(deviance, digits)
        result += "penalty: {:.{}f}\n".format(psum, digits)
        result += "Penalized deviance: {:.{}f}".format(deviance + psum, digits)

        return result

    def __sub__(self, other):
        if not isinstance(other, DIC):
            raise TypeError('The second object must be of type DIC.')

        if self.type != other.type:
            raise ValueError("incompatible dic object: different penalty types")

        delta = self.deviance + self.penalty - \
                other.deviance - other.penalty

        return DiffDIC(delta)

    def __str__(self):
        return self.construct_report()

    def __repr__(self):
        return self.__str__()


def dic_samples(model,
                n_iter,
                thin = 1,
                type = "pD"):
    """
    This function draws samples from a model and computes the
    Deviance Information Criterion (DIC).

    Parameters
    ----------
    model: a PyJAGS model instance
    n_iter: the number of iterations to sample for
    thin: a positive integer specifying thinning interval
    type: either 'pD' or 'popt'

    Returns
    -------
    the Deviance Information Criterion (DIC)
    """
    if not isinstance(model, Model):
        raise ValueError("Invalid JAGS model")

    if model.chains == 1:
        raise ValueError("2 or more parallel chains required")

    if not isinstance(n_iter, int) or n_iter <= 0:
        raise ValueError("n_iter must be a positive integer")

    load_module(name='dic')

    if not type == "pD" or type == "popt":
        raise ValueError(f"type must either be pD or popt but is {type}")
    pdtype = type
    model.console.setMonitors(names=("deviance", pdtype),
                              thin=thin,
                              type="mean")

    model.update(iterations=n_iter)

    # this returns a dictionary
    dev = model.console.dumpMonitors(type="mean", flat=True)

    model.console.clearMonitor(name="deviance", type="mean")
    model.console.clearMonitor(name=pdtype, type="mean")

    return DIC(deviance=dev['deviance'],
               penalty=dev[type],
               type=type)



# Copyright (C) 2016 Tomasz Miasko
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import unittest

import pyjags


class TestModules(unittest.TestCase):

    def test_get_modules_dir(self):
        self.assertIsNotNone(pyjags.get_modules_dir())

    def test_module_loading(self):
        pyjags.load_module('basemod')
        pyjags.load_module('bugs')
        pyjags.load_module('lecuyer')

        self.assertEqual(
            ['basemod', 'bugs', 'lecuyer'],
            pyjags.list_modules())

if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2015-2016 Tomasz Miasko
#               2020 Michael Nowotny
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import deepdish as dd
import numpy as np
import typing as tp


def save_samples_dictionary_to_file(
        filename: str,
        samples: tp.Dict[str, np.ndarray],
        compression: bool = True):
    """
    This function saves a python dictionary of samples that has been generated
    by sample method of a model instance to an HDF5 file.

    Parameters
    ----------
    filename: the path the HDF5 file should be saved to
    samples: a dictionary mapping variable names to Numpy arrays with shape
             (parameter_dimension, chain_length, number_of_chains)
    compression: boolean indicating whether or not to use data compression

    """
    if compression:
        dd.io.save(filename, samples, compression='blosc')
    else:
        dd.io.save(filename, samples, compression=None)


def load_samples_dictionary_from_file(filename: str) \
        -> tp.Dict[str, np.ndarray]:
    """
    This function restores a saved a python dictionary of samples that has been
    generated by sample method of a model instance from an HDF5 file.

    Parameters
    ----------
    filename: the path the HDF5 file is located at

    Returns
    -------
    a dictionary mapping variable names to Numpy arrays with shape
             (parameter_dimension, chain_length, number_of_chains)
    """
    return dd.io.load(filename)



# Copyright (C) 2015-2016 Tomasz Miasko
#               2020 Michael Nowotny
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from .adaptive_sampling import (
    MinimumEffectiveSampleSizeCriterion,
    sample_until
)

from .arviz import convert_pyjags_samples_dict_to_arviz_inference_data
from .chain_utilities import (
    discard_burn_in_samples,
    extract_final_iteration_from_samples_for_initialization,
    merge_parallel_chains,
    merge_consecutive_chains
)

from .dic import dic_samples
from .io import (
    load_samples_dictionary_from_file,
    save_samples_dictionary_to_file
)

from .model import *
from .modules import *
from ._version import get_versions

__version__ = get_versions()['version']
del get_versions



# Copyright (C) 2016 Tomasz Miasko
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

__all__ = ['const_time_partition', 'progressbar']

import math
import sys
import threading
import time
from datetime import timedelta
from functools import wraps


default_timer = getattr(time, 'monotonic', time.time)


def synchronized(func):
    lock = threading.Lock()
    @wraps(func)
    def inner(*args, **kwargs):
        with lock:
            func(*args, **kwargs)
    return inner


def const_time_partition(iterations, period, timer=default_timer):
    """
    Divides iterations into roughly constant time sub-iterations. Time
    necessary to complete a single iteration is estimated as elapsed time
    divided by all already completed iterations.

    Parameters
    ----------
    steps : int
        A non-negative integer specifying total number of steps to execute.
    period : float
        A positive float number describing desired period between yields from
        generator.
    timer : callable, optional
        Monotonic clock, i.e., function returning number of elapsed seconds
        since some arbitrary point in time. Uses ``time.monotonic`` by default,
        if not available falls back to ``time.time``.

    Examples
    --------

    Following example demonstrates how to display information about progress,
    roughly every 5 seconds:

    >>> for steps in const_time_partition(20, 5.0):
    ...     for step in range(steps):
    ...         print('Working')
    ...         time.sleep(1.0)
    ...     print('Progress')
    ...

    """
    start = timer()
    left = iterations
    next = 1
    while left > 0:
        yield next
        elapsed = timer() - start
        left -= next
        done = iterations - left
        next = int(period * done / elapsed) if elapsed > 0 else 2 * next
        if next < 1:
            next = 1
        if next > left:
            next = left


class EmptyProgressBar:

    def __init__(self, *args, **kwargs):
        pass

    def update(self, n):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class ProgressBar:

    FORMAT = 'iterations {self.iterations_done} ' \
             'of {self.iterations_total}, ' \
             'elapsed {self.elapsed}, ' \
             'remaining {self.remaining}'

    def __init__(self, steps, header='', refresh_seconds=0.5,
                 file=sys.stdout, timer=default_timer):
        self.format = header + self.FORMAT
        self.file = file
        self.isatty = file.isatty()
        self.timer = timer
        self.start_seconds = self.timer()
        self.last_seconds = self.start_seconds
        self.refresh_seconds = refresh_seconds
        self.iterations_done = 0
        self.iterations_total = steps
        self.previous_length = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.update(0, force=True)
        if self.isatty:
            self.file.write('\n')
            self.file.flush()

    @synchronized
    def update(self, steps, force=False):
        self.iterations_done += steps
        seconds = self.timer()
        if self.refresh_seconds <= seconds - self.last_seconds or force:
            self.last_seconds = seconds
            self.write(self.render())

    def render(self):
        return self.format.format(self=self)

    def write(self, line):
        if self.isatty:
            # 1. Move to the beginning of the line
            # 2. Overwrite previous content (necessary when new line is shorter)
            # 3. Move to the beginning again.
            n = self.previous_length
            self.file.write('\b' * n + ' ' * n + '\b' * n)
            self.file.write(line)
            self.previous_length = len(line)
        else:
            self.file.write(line)
            self.file.write('\n')
        self.file.flush()

    @property
    def iterations_remaining(self):
        return self.iterations_total - self.iterations_done

    @property
    def percentage(self):
        if self.iterations_total:
            return 100 * self.iterations_done / self.iterations_total
        else:
            return 100

    @property
    def elapsed(self):
        elapsed_seconds = self.last_seconds - self.start_seconds
        return timedelta(seconds=round(elapsed_seconds, 0))

    @property
    def time_per_iteration(self):
        elapsed_seconds = self.last_seconds - self.start_seconds
        return elapsed_seconds / self.iterations_done if self.iterations_done else float('Inf')

    @property
    def remaining(self):
        remaining_seconds = self.iterations_remaining * self.time_per_iteration
        if math.isinf(remaining_seconds):
            return timedelta.max
        else:
            return timedelta(seconds=round(remaining_seconds, 0))


def progress_bar_factory(enable, *args, **kwargs):
    type = ProgressBar if enable else EmptyProgressBar
    def factory(steps, *fargs, **fkwargs):
        all_args = fargs + args
        all_kwargs = dict(kwargs)
        all_kwargs.update(fkwargs)
        return type(steps, *all_args, **all_kwargs)
    return factory



# Copyright (C) 2015-2016 Tomasz Miasko
#               2020 Michael Nowotny
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import numbers
import numpy as np
import typing as tp


def get_chain_length(samples: tp.Dict[str, np.ndarray]) -> int:
    """
    This function determines the length of the chains in the samples dictionary

    Parameters
    ----------
    samples: a dictionary mapping variable names to Numpy arrays with shape
             (parameter_dimension, chain_length, number_of_chains)

    Returns
    -------
    the chain length

    """
    chain_lengths = set(value.shape[1] for key, value in samples.items())

    if samples is None or len(samples) == 0:
        raise ValueError('The samples object must not be empty')

    if len(chain_lengths) > 1:
        raise ValueError(
            'The chain lengths are not consistent across variables.')

    return next(iter(chain_lengths))


def discard_burn_in_samples(
        samples: tp.Dict[str, np.ndarray],
        burn_in: int) -> tp.Dict[str, np.ndarray]:
    """
    This function discards a given number of samples from the beginning of each
    chain for each variable and returns the remaining samples.

    Parameters
    ----------
    samples: a dictionary mapping variable names to Numpy arrays with shape
             (parameter_dimension, chain_length, number_of_chains)

    burn_in: the number of observations to discard from the beginning

    Returns
    -------
    a dictionary with the remaining samples

    """
    return {variable_name: sample_chain[:, burn_in:, :]
            for variable_name, sample_chain
            in samples.items()}


def extract_final_iteration_from_samples_for_initialization(
        samples: tp.Dict[str, np.ndarray],
        variable_names: tp.Set[str]) \
        -> tp.List[tp.Dict[str, tp.Union[numbers.Number, np.ndarray]]]:
    """
    This function extracts the last iteration from each chain for a given set
    of variables.

    Parameters
    ----------
    samples: a dictionary mapping variable names to Numpy arrays with shape
             (parameter_dimension, chain_length, number_of_chains)

    variable_names: a set of variable names

    Returns
    -------
    a dictionary mapping variable names to a numpy array of final samples

    """
    numbers_of_chains = [samples[variable_name].shape[2]
                         for variable_name
                         in variable_names]

    if any(number_of_chains != numbers_of_chains[0] for number_of_chains in
           numbers_of_chains):
        raise ValueError(
            'The number of chains must be identical across parameters')

    number_of_chains = numbers_of_chains[0]

    result = []

    for chain in range(number_of_chains):
        init_chain = {}
        result.append(init_chain)
        for variable_name in variable_names:
            init_chain[variable_name] = \
                samples[variable_name][:, -1, chain].squeeze()

    return result


def _check_sequence_of_chains_present(
        sequence_of_chains: tp.Sequence[tp.Dict[str, np.ndarray]]):
    """
    This function verifies that e sequence of samples is not empty not None.

    Parameters
    ----------
    sequence_of_chains: a sequence of sample dictionaries

    Returns
    -------

    """
    if sequence_of_chains is None:
        raise ValueError('sequence_of_chains must not be none')

    if len(sequence_of_chains) == 0:
        raise ValueError('sequence_of_chains must contain at least one chain')


def _verify_and_get_variable_names_from_sequence_of_samples(
        sequence_of_samples: tp.Sequence[tp.Dict[str, np.ndarray]]) \
        -> tp.Set[str]:
    """
    This function verifies that all sample dictionaries in a sequence contain
    the same set of variables and returns this set of variables.

    Parameters
    ----------
    sequence_of_samples: a sequence of sample dictionaries

    Returns
    -------

    """
    sequence_of_variable_name_sets = \
        [set(sample_chain.keys())
         for sample_chain
         in sequence_of_samples]

    for variable_names in sequence_of_variable_name_sets:
        if variable_names != sequence_of_variable_name_sets[0]:
            raise ValueError('Each sample dictionary must contain the same set '
                             'of variables.')

    return sequence_of_variable_name_sets[0]


def merge_consecutive_chains(
        sequence_of_samples: tp.Sequence[tp.Dict[str, np.ndarray]]) \
        -> tp.Dict[str, np.ndarray]:
    """
    This function concatenates the chains in sample dictionaries sequentially
    (i.e. continues the chains).
    This is useful if samples have been drawn from JAGS consecutively where each
    new sample chain starts from the last iteration of the previous sample.

    Parameters
    ----------
    sequence_of_samples: a sequence of sample dictionaries

    Returns
    -------
    a single sample dictionary merged along chain iterations
    """

    _check_sequence_of_chains_present(sequence_of_samples)

    merged_samples = {}

    variable_names = _verify_and_get_variable_names_from_sequence_of_samples(sequence_of_samples)

    for variable_name in variable_names:
        sequence_of_shapes = \
            [sample_chains[variable_name].shape
             for sample_chains
             in sequence_of_samples]

        sequence_of_numpy_arrays = \
            [sample_chains[variable_name]
             for sample_chains
             in sequence_of_samples]

        parameter_dimension, _, number_of_chains = sequence_of_shapes[0]

        if not all(shape[0] == parameter_dimension
                   for shape
                   in sequence_of_shapes):
            raise ValueError(f'The dimension of {variable_name} is inconsistent'
                             f' between samples.')

        if not all(shape[2] == number_of_chains
                   for shape
                   in sequence_of_shapes):
            raise ValueError('The number of chains is inconsistent across '
                             'samples.')

        merged_samples[variable_name] = \
            np.concatenate(sequence_of_numpy_arrays, axis=1)

    return merged_samples


def merge_parallel_chains(
        sequence_of_samples: tp.Sequence[tp.Dict[str, np.ndarray]]) \
        -> tp.Dict[str, np.ndarray]:
    """
    This function concatenates sample dictionaries across chains
    (i.e. adds more chains of the same length for the same variables).
    This is useful if the starting value of each set of samples is different
    from the last iteration of the chains in the previous sample.

    Parameters
    ----------
    sequence_of_samples: a sequence of sample dictionaries

    Returns
    -------
    a single sample dictionary merged across chains
    """
    _check_sequence_of_chains_present(sequence_of_samples)

    merged_samples = {}

    variable_names = _verify_and_get_variable_names_from_sequence_of_samples(sequence_of_samples)

    for variable_name in variable_names:
        sequence_of_shapes = \
            [sample_chains[variable_name].shape
             for sample_chains
             in sequence_of_samples]

        sequence_of_numpy_arrays = \
            [sample_chains[variable_name]
             for sample_chains
             in sequence_of_samples]

        parameter_dimension, chain_length, _ = sequence_of_shapes[0]

        if not all(shape[0] == parameter_dimension
                   for shape
                   in sequence_of_shapes):
            raise ValueError(f'The dimension of {variable_name} is inconsistent'
                             f' between samples.')

        if not all(shape[1] == chain_length for shape in sequence_of_shapes):
            raise ValueError('The chain lengths are inconsistent across '
                             'samples.')

        merged_samples[variable_name] = \
            np.concatenate(sequence_of_numpy_arrays, axis=2)

    return merged_samples



# Copyright (C) 2015-2016 Tomasz Miasko
#               2020 Michael Nowotny
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import arviz as az
import numpy as np
import typing as tp

from .arviz import convert_pyjags_samples_dict_to_arviz_inference_data
from .chain_utilities import (
    merge_consecutive_chains,
    get_chain_length)

from .model import Model


class MinimumEffectiveSampleSizeCriterion:
    def __init__(self,
                 minimum_ess: int,
                 variable_names: tp.Optional[tp.List[str]] = None):
        """
        This class implements a minimum effective sample size criterion to be
        used with sample_until.

        Parameters
        ----------
        minimum_ess: the minimum effective sample size required
        variable_names: the names of the variables to consider
        """
        self._minimum_ess = minimum_ess
        self._variable_names = variable_names

    @property
    def variable_names(self) -> tp.Optional[tp.List[str]]:
        return self._variable_names

    @property
    def minimum_ess(self) -> int:
        return self._minimum_ess

    def __call__(self,
                 samples: tp.Dict[str, np.ndarray],
                 verbose: bool,
                 ) -> bool:
        print(f'chain_length in __call__ = {get_chain_length(samples)}')
        idata = convert_pyjags_samples_dict_to_arviz_inference_data(samples)
        ess = az.ess(idata, var_names=self.variable_names)

        minimum_ess = min(value['data']
                          for key, value
                          in ess.to_dict()['data_vars'].items())

        if verbose:
            print(f'minimum ess = {minimum_ess}')

        return minimum_ess >= self.minimum_ess


def sample_until(model: Model,
                 criterion: tp.Callable[[tp.Dict[str, np.ndarray], bool], bool],
                 previous_samples: tp.Optional[tp.Dict[str, np.ndarray]] = None,
                 chunk_size: int = 5000,
                 max_iterations: int = 250000,
                 vars: tp.Sequence[str] = None,
                 thin: int = 1,
                 monitor_type: str = "trace",
                 verbose: bool = False) -> tp.Dict[str, np.ndarray]:
    """
    This function progressively samples from a model until a criterion is met.

    Parameters
    ----------
    model: a PyJAGS model
    criterion: a function evaluating a samples dictionary and returning a bool
    previous_samples: an existing sample dictionary to incorporate
    chunk_size: the number of iterations to sample each step
    max_iterations: the maximum number of iterations to sample
    vars: a list of variables to monitor
    thin: a positive integer specifying thinning interval
    monitor_type
    verbose: whether to output step information

    Returns
    -------

    """

    if chunk_size > max_iterations:
        raise ValueError('chunk_size must be less than or equal to '
                         'max_iterations')

    if previous_samples is not None:
        print(f'chain_length at the beginning of sample_until = '
              f'{get_chain_length(previous_samples)}')

    if previous_samples is not None and criterion(previous_samples, verbose):
        return previous_samples

    iterations_left = max_iterations
    while True:
        iterations = min(iterations_left, chunk_size)

        new_samples = model.sample(iterations=iterations,
                                   vars=vars,
                                   thin=thin,
                                   monitor_type=monitor_type)

        if previous_samples is None:
            previous_samples = new_samples
        else:
            previous_samples = \
                merge_consecutive_chains((previous_samples, new_samples))
            print(f'chain_length at the after merging in sample_until = '
                  f'{get_chain_length(previous_samples)}')

        iterations_left -= iterations

        criterion_satisfied = criterion(previous_samples, verbose)

        if criterion_satisfied:
            break
        elif iterations_left <= 1:
            print('maximum number of iterations reached without '
                  'satisfying the criterion')
            break

    return previous_samples



# Copyright (C) 2015-2016 Tomasz Miasko
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import os.path
import sys
import unittest

import numpy as np

import pyjags


class ParametrizedTestCase(unittest.TestCase):

    def __init__(self, method_name, **params):
        super(ParametrizedTestCase, self).__init__(method_name)
        self.__dict__.update(params)


class TestModel(unittest.TestCase):

    def model(self, *args, **kwargs):
        """Create new model instance."""
        return pyjags.Model(*args, **kwargs)

    def test_creating_model_from_string(self):
        # No exceptions should be thrown.
        self.model(code='model { y ~ dbern(1) }')
        self.model(code=b'model { y ~ dbern(1) }')

    def test_creating_model_from_file(self):
        path = os.path.dirname(__file__)
        path = os.path.join(path, 'model.jags')
        self.model(file=path)

    def test_model_is_required(self):
        with self.assertRaises(ValueError):
            self.model()

    def test_random_number_generator_seed(self):
        code = '''
        model {
            x ~ dbern(0.5)
        }
        '''
        init = {
            '.RNG.name': 'base::Wichmann-Hill',
            '.RNG.seed': 1
        }
        n = 100
        s1 = self.model(code, init=init).sample(n)
        s2 = self.model(code, init=init).sample(n)
        np.testing.assert_equal(s1, s2,
                                'Using seed should be give deterministic samples.')

    def test_selecting_random_number_generators(self):
        expected_names = [
            'base::Marsaglia-Multicarry',
            'base::Marsaglia-Multicarry',
            'base::Mersenne-Twister',
            'base::Mersenne-Twister',
            'base::Wichmann-Hill',
        ]
        code = 'model { x ~ dbern(0.5) }'
        init = [{'.RNG.name': name, '.RNG.seed': 0} for name in expected_names]
        chains = len(init)

        model = self.model(code, init=init, chains=chains, adapt=10)
        model.sample(20)

        actual_names = [v for state in model.state
                        for k, v in state.items()
                        if k == '.RNG.name']
        self.assertEqual(expected_names, actual_names)

    def test_empty_array(self):
        # This used to throw an exception.
        code = 'model { x ~ dbern(1) }'
        data = {'x': []}
        self.model(code, data=data)

    def test_invalid_length_of_initial_value_list_throws_exception(self):
        model = 'model { x ~ dbern(1) }'
        with self.assertRaises(ValueError):
            self.model(model, chains=2, init=[dict(x=1)])

    def test_invalid_type_of_init_value_list(self):
        code = 'model { x ~ dbern(1) }'
        with self.assertRaises(ValueError):
            self.model(code, chains=2, init=1234)

    def test_model_variables(self):
        code = '''
        data {
            a <- 1
        }
        model {
            b <- 2
            for (i in 1:3) {
                x[i] ~ dbeta(1, 1)
            }
        }
        '''

        model = self.model(code)
        self.assertEqual({'a', 'b', 'x'}, set(model.variables))

    def test_model_data(self):
        code = '''
        model {
            for (i in 1:N) {
                x[i] ~ dbin(p, n[i])
            }
            p ~ dbeta(1, 1)
        }
        '''

        N = 100
        n = np.random.random_integers(1, 10, N)
        x = np.random.binomial(n, 0.10, N)
        x = np.ma.masked_array(x, np.random.choice([0, 1], size=N))
        data = dict(n=n, x=x, N=N)

        m = self.model(code, data=data)
        model_data = m.data

        self.assertEqual(data.keys(), model_data.keys())
        for var in data.keys():
            np.testing.assert_equal(data[var], model_data[var])

    def test_model_parameters(self):
        code = '''
        model {
            for (i in 1:10) {
                x[i] ~ dnorm(mu[i], 1)
                mu[i] ~ dunif(-1, 1)
            }
        }
        '''
        chains = 3
        model = self.model(code, data=dict(x=np.zeros(10)), chains=chains)

        parameters = model.parameters
        self.assertEqual(chains, len(parameters))
        names =  set(parameters[0].keys())
        self.assertEqual({'mu', '.RNG.name', '.RNG.state'}, names)

    def test_samples_shape(self):
        code = '''
        model {
            for (i in 1:3) {
                for (j in 1:5) {
                    x[i, j] ~ dnorm(mu[i], 1)

                }
                mu[i] ~ dunif(-1, 1)
            }
        }
        '''

        chains = 7
        iterations = 17
        data = {'x': np.zeros((3, 5))}

        m = self.model(code, data=data, chains=chains)
        s = m.sample(iterations)

        self.assertEqual(s['x'].shape, (3, 5, iterations, chains))
        self.assertEqual(s['mu'].shape, (3, iterations, chains))

    def test_missing_input_data(self):
        code = '''
        model {
            for (i in 1:length(x)) {
                x[i] ~ dbern(0.5)
            }
        }'''

        data = {'x': np.ma.masked_outside([0, 1, -1], 0, 1)}
        c = 2
        m = self.model(code, data=data, chains=c)
        n = 100
        s = m.sample(n, vars=['x'])

        x1 = s['x'][0,:,:]
        x2 = s['x'][1,:,:]
        x3 = s['x'][2,:,:]

        # Observed values, samples should be constant
        np.testing.assert_equal([0] * n * c, x1.flatten())
        np.testing.assert_equal([1] * n * c, x2.flatten())
        # Missing value, samples should vary between 0 and 1
        self.assertIn(0, x3)
        self.assertIn(1, x3)

    @unittest.skipIf(pyjags.version() < (4,0,0), "Not supported before JAGS 4.0.0")
    def test_missing_sample_data(self):
        code = '''
        model {
            x[1] ~ dnorm(0, 10)
            x[3] ~ dnorm(0, 15)
        }'''

        m = self.model(code, chains=2)
        s = m.sample(10, vars=['x'])

        x = s['x']
        x1 = x[0]
        x2 = x[1]
        x3 = x[2]

        # x[2] should be completely masked
        self.assertTrue(np.all(x2.mask))
        # x[1] and x[3] should be plain unmasked numpy arrays
        self.assertFalse(np.ma.is_mask(x1))
        self.assertFalse(np.ma.is_mask(x3))

    def test_unused_variables_throws_exception(self):
        code = 'model { x ~ dbern(0.5) }'

        with self.assertRaises(ValueError):
            self.model(code, data=dict(x=1, y=2))

        with self.assertRaises(ValueError):
            self.model(code, init=dict(x=1, y=2))


class TestModelWithoutProgressBar(TestModel):
    def model(self, *args, **kwargs):
        return pyjags.Model(*args, progress_bar=False, **kwargs)


if sys.version_info[0] > 2:

    class TestModelWithThreads(TestModel):

        def model(self, *args, **kwargs):
            return pyjags.Model(*args, threads=3, **kwargs)


    class TestModelWithChainsPerThread(TestModel):

        def model(self, *args, **kwargs):
            return pyjags.Model(*args, threads=3, chains_per_thread=2, **kwargs)

if __name__ == '__main__':
    unittest.main()



# encoding: utf-8
#
# Copyright (C) 2015-2016 Tomasz Miasko
#               2020 Michael Nowotny
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import sys

from setuptools import dist, setup, Extension
import subprocess
import versioneer


def content(path):
    with open(path) as fh:
        return fh.read()


def add_pkg_config(ext, package):
    flags_map = {
        '-I': ['include_dirs'],
        '-L': ['library_dirs', 'runtime_library_dirs'],
        '-l': ['libraries'],
    }

    try:
        args = ['pkg-config', '--libs', '--cflags', package]
        output = subprocess.check_output(args)
        output = output.decode()
        for flag in output.split():
            for attr in flags_map[flag[:2]]:
                getattr(ext, attr).append(flag[2:])

        args = ['pkg-config', '--modversion', package]
        output = subprocess.check_output(args, universal_newlines=True)
        return output.strip()
    except Exception as err:
        print("Error while executing pkg-config: {}".format(err))
        sys.exit(1)


def add_jags(ext):
    version = add_pkg_config(ext, 'jags')
    version = '"{}"'.format(version)
    ext.define_macros.append(('PYJAGS_JAGS_VERSION', version))


def add_numpy(ext):
    try:
        import numpy
    except ImportError:
        sys.exit('Please install numpy first.')
    ext.include_dirs.append(numpy.get_include())


def add_pybind11(ext):
    ext.include_dirs.append('pybind11/include')
    ext.extra_compile_args.append('-std=c++17')


if __name__ == '__main__':
    ext = Extension('pyjags.console',
                    language='c++',
                    sources=['pyjags/console.cc'])
    add_jags(ext)
    add_numpy(ext)
    add_pybind11(ext)

    setup(name='pyjags',
          version=versioneer.get_version(),
          cmdclass=versioneer.get_cmdclass(),
          description='Python interface to JAGS library for Bayesian data analysis.',
          long_description=content('README.rst'),
          author=u'Tomasz Miąsko, Michael Nowotny',
          author_email='tomasz.miasko@gmail.com, nowotnym@gmail.com',
          url='https://github.com/michaelnowotny/pyjags',
          license='GPLv2',
          classifiers=[
              'Development Status :: 4 - Beta',
              'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
              'Operating System :: POSIX',
              'Programming Language :: C++',
              'Programming Language :: Python :: 2',
              'Programming Language :: Python :: 2.7',
              'Programming Language :: Python :: 3',
              'Programming Language :: Python :: 3.4',
              'Programming Language :: Python :: 3.5',
              'Programming Language :: Python :: 3.6',
              'Programming Language :: Python :: 3.7',
              'Programming Language :: Python :: 3.8',
              'Programming Language :: Python',
              'Topic :: Scientific/Engineering :: Information Analysis',
              'Topic :: Scientific/Engineering',
          ],
          packages=['pyjags'],
          ext_modules=[ext],
          install_requires=['arviz',
                            'deepdish',
                            'numpy'],
          test_suite='test')



# Copyright (C) 2015-2016 Tomasz Miasko
#               2020 Michael Nowotny
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import arviz as az
import numpy as np
import typing as tp


def convert_pyjags_samples_dict_to_arviz_inference_data(
        samples: tp.Dict[str, np.ndarray]) -> az.InferenceData:
    """
    This function takes a python dictionary of samples that has been generated
    by sample method of a model instance and returns an Arviz inference data
    object.
    Parameters
    ----------
    samples: a dictionary mapping variable names to Numpy arrays with shape
             (parameter_dimension, chain_length, number_of_chains)

    Returns
    -------
    An Arviz inference data object
    """
    # pyjags returns a dictionary of numpy arrays with shape
    #         (parameter_dimension, chain_length, number_of_chains)
    # but arviz expects samples with shape
    #         (number_of_chains, chain_length, parameter_dimension)

    parameter_name_to_samples_map = {}

    for parameter_name, chains in samples.items():
        parameter_dimension, chain_length, number_of_chains = chains.shape
        if parameter_dimension == 1:
            parameter_name_to_samples_map[parameter_name] = \
                chains[0, :, :].transpose()
        else:
            for i in range(parameter_dimension):
                parameter_name_to_samples_map[f'{parameter_name}_{i+1}'] = \
                    chains[i, :, :].transpose()

    return az.InferenceData(
        posterior=az.data.base.dict_to_dataset(parameter_name_to_samples_map))
